import base64
import hashlib
import re
import string
import unicodedata
from typing import NamedTuple, Dict, List, Set, Iterable, Optional
from urllib.parse import quote_plus, urlencode

from sqlalchemy.engine import Connection, Dialect

from bach.expression import Expression, ColumnReferenceToken
from bach.sql_model import BachSqlModel
from sql_models.util import is_postgres, DatabaseNotSupportedException, is_bigquery, is_athena


class SortColumn(NamedTuple):
    expression: Expression
    asc: bool


class FeatureRange(NamedTuple):
    min: int
    max: int


class ResultSeries(NamedTuple):
    name: str
    expression: 'Expression'
    dtype: str


def get_result_series_dtype_mapping(result_series: List[ResultSeries]) -> Dict[str, str]:
    return {
        rs.name: rs.dtype
        for rs in result_series
    }


def get_merged_series_dtype(dtypes: Set[str]) -> str:
    """
    returns a final dtype when trying to combine series with different dtypes
    """
    from bach import get_series_type_from_dtype, SeriesAbstractNumeric
    if len(dtypes) == 1:
        return dtypes.pop()
    elif all(
        issubclass(get_series_type_from_dtype(dtype), SeriesAbstractNumeric)
        for dtype in dtypes
    ):
        return 'float64'

    # default casting will be as text, this way we avoid any SQL errors
    # when merging different db types into a column
    return 'string'


def escape_parameter_characters(conn: Connection, raw_sql: str) -> str:
    """
    Return a modified copy of the given sql with the query-parameter special characters escaped.
    e.g. if the connection uses '%' to mark a parameter, then all occurrences of '%' will be replaced by '%%'
    """
    # for now we'll just assume Postgres and assume the pyformat parameter style is used.
    # When we support more databases we'll need to do something smarter, see
    # https://www.python.org/dev/peps/pep-0249/#paramstyle
    return raw_sql.replace('%', '%%')


_BQ_COLUMN_NAME_RESERVED_PREFIXES = [
    # https://cloud.google.com/bigquery/docs/schemas#column_names
    '_TABLE_',
    '_FILE_',
    '_PARTITION',
    '_ROW_TIMESTAMP',
    '__ROOT__',
    '_COLIDENTIFIER'
]


def is_valid_column_name(dialect: Dialect, name: str) -> bool:
    """
    Check that the given name is a valid column name in the SQL dialect.
    """
    if is_postgres(dialect):
        # Identifiers longer than 63 characters are not necessarily wrong, but they will be truncated which
        # could lead to identifier collisions, so we just disallow it.
        # source: https://www.postgresql.org/docs/14/sql-syntax-lexical.html#SQL-SYNTAX-IDENTIFIERS
        return 0 < len(name) < 64
    if is_athena(dialect):
        # Source: https://docs.aws.amazon.com/athena/latest/ug/tables-databases-columns-names.html

        # Only allow lower-case a-z, to make sure we don't have duplicate case-insensitive column names
        regex = '^[a-z0-9_]*$'
        len_ok = 0 < len(name) <= 255
        pattern_ok = bool(re.match(pattern=regex, string=name))
        return len_ok and pattern_ok
    if is_bigquery(dialect):
        # Sources:
        #  https://cloud.google.com/bigquery/docs/reference/standard-sql/lexical#column_names
        #  https://cloud.google.com/bigquery/docs/reference/standard-sql/lexical#case_sensitivity
        #  https://cloud.google.com/bigquery/docs/schemas#column_names

        # Only allow lower-case a-z, to make sure we don't have duplicate case-insensitive column names
        regex = '^[a-z_][a-z0-9_]*$'
        len_ok = 0 < len(name) <= 300
        pattern_ok = bool(re.match(pattern=regex, string=name))
        prefix_ok = not any(name.startswith(prefix.lower()) for prefix in _BQ_COLUMN_NAME_RESERVED_PREFIXES)
        return len_ok and pattern_ok and prefix_ok
    raise DatabaseNotSupportedException(dialect)


def get_sql_column_name(dialect: Dialect, name: str) -> str:
    """
    Given the name of a series and dialect, return the sql column name.

    If the name contains no characters that require escaping, for the given dialect, then the same string is
    returned. Otherwise, a column name is that can be safely used with the given dialect. The generated name
    will be based on the given name, but with all non-allowed characters filtered out and a hash added.

    **Background**
    Each Bach Series is mapped to a column in queries. Unfortunately, some databases only supported a limited
    set of characters in column names. To work around this limitation we distinguish the Series name from the
    sql-column name. The former is the name a Bach user will see and use, the latter is the name that will
    appear in SQL queries.

    The algorithm we use to map series names to column names is deterministic, but not reversible. That is,
    the same input always yields the same output, and it might be impossible to recover the input from a
    given output.

    :raises ValueError: if name cannot be appropriately escaped or is too long for the given dialect
    :return: column name for the given series name.
    """
    if is_valid_column_name(dialect, name):
        return name

    # If name is not a valid column name in itself. Then transform it:
    # 1) Turn `name` into `cleaned_name`, which only keeps the allowed characters
    # 2) Add a hash of `name` to make the name unique again
    #
    # This process should yield a valid column name for any database, if it's not too long.

    table_upper_to_lower = str.maketrans(string.ascii_uppercase, string.ascii_lowercase)
    chars_lower = set(string.ascii_lowercase)
    chars_numbers = set('0123456789')
    chars_underscore = set('_')
    allowed_chars = chars_lower | chars_numbers | chars_underscore

    # Convert 'name' into cleaned form, in three steps:
    # 1) Normalize the name to the decomposed character form. This will split diacritics of the base
    #    character, e.g. 'ç' will be split in two characters: 'c' and ' ̧'.
    # 2) Lowercase all characters.
    # 3) Filter out all characters that are not in [a-z0-9_]. Example: 'c' would be kept, but ' ̧' would
    #    be filtered out.

    # There are probably better methods, and libraries that have more comprehensive ways of doing this.
    # But for now this works okayish, and is simple.
    # In step 2 we lowercase characters using a translation table. We could probably just do .lower(), but I
    # wasn't 100% sure that works in all locales the same, and the docs were unclear on that.

    decomposed_characters = unicodedata.normalize('NFKD', name)
    decomposed_lowered = decomposed_characters.translate(table_upper_to_lower)
    cleaned_name = ''.join(c for c in decomposed_lowered if c in allowed_chars)
    if not cleaned_name:
        cleaned_name = 'empty'

    if is_bigquery(dialect):
        # Additional rules for BigQuery:
        # 1) Must start with a character a-z or underscore, not with a number
        if cleaned_name[0] in chars_numbers:
            cleaned_name = f'_{cleaned_name}'
        # 2) Cannot start with one of the reserved prefixes:
        if any(cleaned_name.startswith(prefix.lower()) for prefix in _BQ_COLUMN_NAME_RESERVED_PREFIXES):
            cleaned_name = f'x_{cleaned_name}'

    # Add a hash of `name` to make the name unique again.
    # Ten characters of hash (50 bits) should be enough.
    hash = _get_string_hash(name)[:10]
    escaped = f'{cleaned_name}__{hash}'

    if not is_valid_column_name(dialect, escaped):
        raise ValueError(f'Column name "{name}" is not valid for SQL dialect {dialect.name}, and cannot be '
                         f'escaped. Try making the series name shorter and/or using characters [a-z] only.')
    return escaped


def _get_string_hash(input: str) -> str:
    """ Give md5 hash of the given input strings, encoded as a 26 character long base-32 string. """
    # .hexdigest() would give the md5 hash encoded with [a-e0-9], which gives 4 bits per character
    # we'd rather not spend too many characters on the hash, so we base-32 encode the bytes, which gives 5
    # bits per character. Which makes the 128-bit hash fit in 26 characters instead of 32.
    input_bytes = input.encode('utf-8')
    md5_bytes = hashlib.md5(input_bytes).digest()
    hash = base64.b32encode(md5_bytes).decode('utf-8').lower().replace('=', '')
    return hash


def get_name_to_column_mapping(dialect: Dialect, names: Iterable[str]) -> Dict[str, str]:
    """ Give a mapping of series names to sql column names. """
    return {
        name: get_sql_column_name(dialect=dialect, name=name) for name in names
    }


def validate_node_column_references_in_sorting_expressions(
    dialect: Dialect, node: BachSqlModel, order_by: List[SortColumn],
) -> None:
    """
    Validate that all ColumnReferenceTokens in order_by expressions refer columns that exist in node.
    """
    sql_column_names = set(get_sql_column_name(dialect=dialect, name=column) for column in node.series_names)
    for ob in order_by:
        invalid_column_references = [
            token.column_name
            for token in ob.expression.get_all_tokens()
            if isinstance(token, ColumnReferenceToken) and token.column_name not in sql_column_names
        ]
        if invalid_column_references:
            raise ValueError(
                (
                    'Sorting contains expressions referencing '
                    f'non-existent columns in current base node. {invalid_column_references}.'
                    ' Please call DataFrame.sort_values([]) or DataFrame.sort_index() for removing'
                    ' sorting and try again.'
                )
            )


def merge_sql_statements(dialect: Dialect, sql_statements: List[str]) -> List[str]:
    """
    Merge multiple sql statements into one statement with separating semicolons, if the dialect supports
    executing multiple statements in once call to conn.execute(). Otherwise return the original list.
    """
    if is_athena(dialect) or not sql_statements:
        return sql_statements
    combined = '; '.join(sql_statements)
    return [combined]


def athena_construct_engine_url(
    *,
    aws_access_key_id: str = None,
    aws_secret_access_key: str = None,
    region_name: str,
    schema_name: str,
    s3_staging_dir: str,
    athena_work_group: Optional[str] = None,
    catalog_name: Optional[str] = None
) -> str:
    """
    Construct a url that can be passed to SqlAlchemy's `create_engine()` to connect to an Athena database.

    Note that the access key and secret access key are not needed for a valid url. Athena libraries will use
    the standard fallback mechanism in case credentials are missing from the url, see
    https://boto3.amazonaws.com/v1/documentation/api/latest/guide/credentials.html#configuring-credentials

    For more information on

    :param aws_access_key_id: Optional, account to use
    :param aws_secret_access_key: Optional, secret key for account
    :param region_name: region where data is hosted, e.g. 'eu-west-1'
    :param schema_name: Database name
    :param s3_staging_dir: S3 path where query results should be written,
            e.g. 'S3://some_bucket/my_staging_area/'
    :param athena_work_group: Optional, workgroup under which queries will be executed.
    :param catalog_name: Optional, catalog in which table information is stored.
    """
    q_region_name = quote_plus(region_name)
    q_schema_name = quote_plus(schema_name)

    if aws_access_key_id and aws_secret_access_key:
        q_user_pass = f'{quote_plus(aws_access_key_id)}:{quote_plus(aws_secret_access_key)}@'
    elif aws_access_key_id:
        q_user_pass = f'{quote_plus(aws_access_key_id)}@'
    else:
        q_user_pass = ''

    query_string_params = {'s3_staging_dir': s3_staging_dir}
    if athena_work_group is not None:
        query_string_params['work_group'] = athena_work_group
    if catalog_name:
        query_string_params['catalog_name'] = catalog_name

    query_string = urlencode(query_string_params)

    url = (
        f'awsathena+rest://'
        f'{q_user_pass}athena.{q_region_name}.amazonaws.com:443/{q_schema_name}?{query_string}'
    )
    return url



"""
Copyright 2021 Objectiv B.V.
"""
from typing import NamedTuple, Optional, List

import sqlparse
from sqlparse.sql import TokenList, Token


class CteTuple(NamedTuple):
    name: Optional[str]
    select_sql: str


def raw_sql_to_selects(sql: str) -> List[CteTuple]:
    """
    Given a raw sql select statement, return a list with all CTEs and the final select statement.

    If the sql consists of a simple query without CTEs then the returned list contains a single item.
    If the sql consists of a CTEs followed by a final select, then the returned list contains all CTEs in
    the encountered order, followed by the select statement.
    For each CTE in the returned value the `name` field is guaranteed to be set, for the final select
    statement it will be None.
    """
    # TODO: refactor function
    stmts = sqlparse.parse(sql)
    if len(stmts) == 0:
        return []

    # if len(stmts) > 1
    #     raise ValueError(f'Expected sql with a single statement, found {len(stmts)} statements.')
    # todo: check that there is one real statement. make sure this check works even if the statement
    #  ends with a semicolon

    stmt = stmts[0]
    tokens = list(stmt.flatten())
    token_list = TokenList(tokens)

    idx_with, _ = token_list.token_next_by(m=(sqlparse.tokens.Keyword.CTE, 'WITH'))
    idx_select, _ = token_list.token_next_by(m=(sqlparse.tokens.DML, 'SELECT'))

    if idx_select is None:
        raise ValueError(f'Cannot find select statement. sql: {sql}')
    if idx_with is None or idx_select < idx_with:
        # This query is a simple select statement
        return [CteTuple(name=None, select_sql=str(stmt))]
    else:
        # This query contains Common Table Expressions
        idx = idx_with
        result = []
        while idx < len(token_list.tokens):
            idx_name, token_name = token_list.token_next_by(idx=idx, t=sqlparse.tokens.Name)
            idx_select, _ = token_list.token_next_by(idx=idx, m=(sqlparse.tokens.DML, 'SELECT'))
            if idx_name is None and idx_select is None:
                raise ValueError(f'Cannot find next CTE or select statement. sql: {sql}')
            if idx_name is None or idx_select < idx_name:
                token_sublist = token_list.tokens[idx_select:]
                result.append(CteTuple(name=None, select_sql=_tokens_to_sql(token_sublist)))
                return result

            # todo: support syntax 'cte_name (column_names) as ...` now we only support 'cte_name as ...'
            idx_as, _ = token_list.token_next_by(idx=idx_name, m=(sqlparse.tokens.Keyword, 'AS'))
            idx_paren_open, _ = token_list.token_next_by(idx=idx_as, m=(sqlparse.tokens.Punctuation, '('))
            idx = idx_paren_open + 1
            paren_open = 1
            while idx < len(token_list.tokens):
                if token_list[idx].match(sqlparse.tokens.Punctuation, '('):
                    paren_open += 1
                elif token_list[idx].match(sqlparse.tokens.Punctuation, ')'):
                    paren_open -= 1
                idx += 1
                if paren_open == 0:
                    # Found the end of the CTE
                    token_sublist = token_list.tokens[(idx_paren_open + 1):(idx - 1)]
                    result.append(
                        CteTuple(name=token_name.normalized, select_sql=_tokens_to_sql(token_sublist))
                    )
                    break
    raise ValueError(f'Reached end of sql code without finding a select statement. sql: {sql}')


def _tokens_to_sql(tokens: List[Token]) -> str:
    return str(TokenList(tokens))



"""
Copyright 2021 Objectiv B.V.
"""
import re
from dataclasses import dataclass
from typing import Optional, Union, TYPE_CHECKING, List, Dict, Tuple, Set, Sequence

from sqlalchemy.engine import Dialect

from sql_models.model import escape_raw_sql
from sql_models.util import quote_string, quote_identifier

if TYPE_CHECKING:
    from bach import Series
    from bach.sql_model import BachSqlModel


@dataclass(frozen=True)
class ExpressionToken:
    """ Abstract base class of ExpressionTokens"""

    def __post_init__(self):
        # Make sure that other code can rely on an ExpressionToken always being a subclass of this class.
        if self.__class__ == ExpressionToken:
            raise TypeError("Cannot instantiate ExpressionToken directly. Instantiate a subclass.")

    def to_sql(self, dialect: Dialect):
        """
        Must be implemented by subclasses. Generated SQL must be assumed to be used as raw sql in
        SqlModel.sql, therefore unknown/untrusted/etc. values should be properly escaped:
         * escape_format_string() twice, unless the value is a specific SqlModel placeholder or reference
         * other escaping/quoting as needed.
        """
        # Not abstract so we can stay a dataclass.
        raise NotImplementedError()


@dataclass(frozen=True)
class RawToken(ExpressionToken):
    raw: str

    def to_sql(self, dialect: Dialect) -> str:
        return escape_raw_sql(self.raw)


@dataclass(frozen=True)
class VariableToken(ExpressionToken):
    dtype: str
    name: str

    def to_sql(self, dialect: Dialect) -> str:
        return '{' + self.dtype_name_to_placeholder_name(self.dtype, self.name) + '}'

    @property
    def dtype_name(self):
        from bach.dataframe import DtypeNamePair
        return DtypeNamePair(dtype=self.dtype, name=self.name)

    @classmethod
    def dtype_name_to_placeholder_name(cls, dtype: str, name: str) -> str:
        return f'___bach_variable___{dtype}___{name}'

    @classmethod
    def placeholder_name_to_token(cls, placeholder_name: str) -> Optional['VariableToken']:
        """
        Reverse of dtype_name_to_placeholder_name().
        Will return None if the placeholder_name doesn't match the pattern
        """
        match = re.match('^___bach_variable___([a-zA-Z0-9)]+)___(.+)$', placeholder_name)
        if not match:
            return None
        return cls(match.group(1), match.group(2))


@dataclass(frozen=True)
class TableColumnReferenceToken(ExpressionToken):
    table_name: Optional[str]
    column_name: str

    def to_sql(self, dialect: Dialect):
        t = f'{quote_identifier(dialect, self.table_name)}.' if self.table_name else ''
        col_name = quote_identifier(dialect, self.column_name)
        return escape_raw_sql(f'{t}{col_name}')


@dataclass(frozen=True)
class ColumnReferenceToken(ExpressionToken):
    column_name: str

    def to_sql(self, dialect: Dialect):
        raise ValueError('ColumnReferenceTokens should be resolved first using '
                         'Expression.resolve_column_references')

    def resolve(self, table_name: Optional[str]) -> TableColumnReferenceToken:
        return TableColumnReferenceToken(table_name=table_name, column_name=self.column_name)


@dataclass(frozen=True)
class ModelReferenceToken(ExpressionToken):
    model: 'BachSqlModel'

    def refname(self) -> str:
        return f'reference{self.model.hash}'

    def to_sql(self, dialect: Dialect) -> str:
        return f'{{{{{self.refname()}}}}}'


@dataclass(frozen=True)
class StringValueToken(ExpressionToken):
    """ Wraps a string value. The value in this object is unescaped and unquoted. """
    value: str

    def to_sql(self, dialect: Dialect) -> str:
        return escape_raw_sql(quote_string(dialect, self.value))


@dataclass(frozen=True)
class IdentifierToken(ExpressionToken):
    name: str

    def to_sql(self, dialect: Dialect) -> str:
        return escape_raw_sql(quote_identifier(dialect, self.name))


class Expression:
    """
    Immutable object representing a fragment of SQL as a sequence of sql-tokens or Expressions.

    Expressions can easily be converted to a string with actual sql using the to_sql() function. Storing a
    sql-expression using this class, rather than storing it directly as a string, makes it possible to
    for example substitute the table-name after constructing the expression.
    Additionally this move this burden of correctly quoting and escaping string literals to this class, if
    literals are expressed with the correct tokens at least.
    In the future we might add support for more literal types.

    This class does not offer full-tokenization of sql. There are only a limited number of tokens for the
    needed use-cases. Most sql is simply encoded as a 'raw' token.

    For special type Expressions, this class is subclassed to assign special properties to a subexpression.
    """

    def __init__(self, data: Union['Expression', Sequence[Union[ExpressionToken, 'Expression']]] = None):
        if not data:
            data = []
        if isinstance(data, Expression):
            # if we only got a base Expression, we absorb it.
            data = data.data if type(data) is Expression else [data]
        self._data: Tuple[Union[ExpressionToken, 'Expression'], ...] = tuple(data)

    @property
    def data(self) -> List[Union[ExpressionToken, 'Expression']]:
        return list(self._data)

    def __eq__(self, other):
        return isinstance(other, Expression) and self.data == other.data

    def __repr__(self):
        return f'{self.__class__}({repr(self.data)})'

    def __hash__(self):
        return hash(self._data)

    @classmethod
    def construct(cls, fmt: str, *args: Union['Expression', 'Series']) -> 'Expression':
        """
        Construct an Expression using a format string that can refer existing expressions.
        Every occurrence of `{}` in the fmt string will be replace with a provided expression (in order that
        they are given). All other parts of fmt will be converted to RawTokens.

        As a convenience, instead of Expressions it is also possible to give Series as args, in that
        case the series's expression is taken as Expression.

        :param fmt: format string
        :param args: 0 or more Expressions or Series. Number of args must exactly match number of `{}`
            occurrences in fmt.
        """

        sub_strs = fmt.split('{}')
        data: List[Union[ExpressionToken, Expression]] = []
        if len(args) != len(sub_strs) - 1:
            raise ValueError(f'For each {{}} in the fmt there should be an Expression provided. '
                             f'Found {{}}: {len(sub_strs) - 1}, provided expressions: {len(args)}')
        for i, sub_str in enumerate(sub_strs):
            if i > 0:
                arg = args[i - 1]
                if not isinstance(arg, Expression):  # arg is a Series
                    arg_expr = arg.expression
                else:
                    arg_expr = arg

                if isinstance(arg_expr, NonAtomicExpression):
                    data.extend([RawToken('('), arg_expr, RawToken(')')])
                else:
                    data.append(arg_expr)
            if sub_str != '':
                data.append(RawToken(raw=sub_str))
        return cls(data=data)

    @classmethod
    def construct_expr_as_sql_name(cls, dialect: Dialect, expr: 'Expression', name: str) -> 'Expression':
        """
        Construct an expression that represents the sql: {expr} as "name".
        The name parameter will be run through `get_sql_name()`, which might transform it into another
        name. This is to make sure the generated expression contains a column name that is legal for the
        specified Dialect.
        """
        from bach.utils import get_sql_column_name
        ident_expr = Expression.identifier(get_sql_column_name(dialect=dialect, name=name))
        # TODO: enable this optimisation again
        # if expr.to_sql() == ident_expr.to_sql():
        #     # this is to prevent generating sql of the form `x as x`, we'll just return `x` in that case
        #     return expr
        return cls.construct('{} as {}', expr, ident_expr)

    @classmethod
    def raw(cls, raw: str) -> 'Expression':
        """ Return an expression that contains a single RawToken. """
        return cls([RawToken(raw)])

    @classmethod
    def variable(cls, dtype: str, name: str) -> 'Expression':
        """ Return an expression that contains a single VariableToken. """
        return cls([VariableToken(dtype=dtype, name=name)])

    @classmethod
    def string_value(cls, value: str) -> 'Expression':
        """
        Return an expression that contains a single StringValueToken with the value.
        :param value: unquoted, unescaped string value.
        """
        return cls([StringValueToken(value)])

    @classmethod
    def identifier(cls, name: str) -> 'Expression':
        return cls([IdentifierToken(name=name)])

    @classmethod
    def column_reference(cls, field_name: str) -> 'Expression':
        """ Construct an expression for field-name, where field-name is a column in a table or CTE. """
        return cls([ColumnReferenceToken(field_name)])

    @classmethod
    def table_column_reference(cls, table_name: str, field_name: str) -> 'Expression':
        """ Construct an expression for table referenced field,
         where table_name is a reference of a table or CTE from which field_name is a column """
        return cls([TableColumnReferenceToken(table_name, field_name)])

    @classmethod
    def model_reference(cls, model: 'BachSqlModel') -> 'Expression':
        """ Construct an expression for model, where model is a reference to a model. """
        return cls([ModelReferenceToken(model)])

    @property
    def is_single_value(self):
        """
        Will this expression return just one value (at most)

        Any Expression made up out of Tokens and Expressions, where all Expressions are single values,
        are expected to also yield a single value. Leaves consisting only of Tokens are considered
        not single valued, so at least one SingleValueExpression need to be present for a branch to
        become single valued.
        """
        if isinstance(self, SingleValueExpression):
            return True
        all_single_value = [d.is_single_value for d in self._data if isinstance(d, Expression)]
        return len(all_single_value) and all(all_single_value)

    @property
    def is_constant(self):
        """
        Does this expression represent a constant value, or an expressions constructed of only constants

        Any Expression made up out of Tokens and Expressions, where all Expressions are constant
        is considered constant. Leaves consisting only of Tokens are considered not constant, so
        at least one ConstValueExpressions need to be present for a branch to become constant.
        """
        if isinstance(self, ConstValueExpression):
            return True
        all_constant = [d.is_constant for d in self._data if isinstance(d, Expression)]
        return len(all_constant) and all(all_constant)

    @property
    def is_independent_subquery(self):
        return isinstance(self, IndependentSubqueryExpression)

    @property
    def has_aggregate_function(self) -> bool:
        """
        True iff we are a AggregateFunctionExpression, or there is at least one in this Expression.
        """
        return isinstance(self, AggregateFunctionExpression) or any(
            d.has_aggregate_function for d in self.data if isinstance(d, Expression)
        )

    @property
    def has_windowed_aggregate_function(self) -> bool:
        """
        True iff we are a WindowFunctionExpression, or there is at least one in this Expression.
        """
        return isinstance(self, WindowFunctionExpression) or any(
            d.has_windowed_aggregate_function for d in self.data if isinstance(d, Expression)
        )

    @property
    def has_table_column_references(self) -> bool:
        """
        True iff we are a TableColumnReference, or there is at least one in this Expression.
        """
        return any(
            isinstance(token, TableColumnReferenceToken) for token in self.get_all_tokens()
        )

    @property
    def has_multi_level_expressions(self) -> bool:
        return isinstance(self, MultiLevelExpression) or any(
            d.has_multi_level_expressions for d in self.data if isinstance(d, Expression)
        )

    def resolve_column_references(self, dialect: Dialect, table_name: Optional[str]) -> 'Expression':
        """ resolve the table name aliases for all columns in this expression """
        result: List[Union[ExpressionToken, Expression]] = []
        for data_item in self.data:
            if isinstance(data_item, Expression):
                result.append(data_item.resolve_column_references(dialect, table_name))
            elif isinstance(data_item, ColumnReferenceToken):
                result.append(data_item.resolve(table_name))
            else:
                result.append(data_item)
        return self.__class__(result)

    def replace_column_references(self, old_column_name: str, new_column_name: str) -> 'Expression':
        """
        replaces all ColumnReferenceToken where old_column_name is present with another ColumnReferenceToken
        """
        replaced_tokens = []
        for token in self.get_all_tokens():
            if not isinstance(token, ColumnReferenceToken) or token.column_name != old_column_name:
                replaced_tokens.append(token)
                continue
            replaced_tokens.append(ColumnReferenceToken(new_column_name))
        return self.__class__(replaced_tokens)

    def remove_table_column_references(self) -> Tuple[Optional[str], Optional[str], 'Expression']:
        """
        removes all table references from this expression.
        Returns first table_name and column_name found and a new expression without table column references.
        If no TableColumnReferenceToken are found, then the table_name and column_name will be None.
        """
        table_name = None
        column_name = None
        if not self.has_table_column_references:
            return table_name, column_name, self

        new_tokens: List[Union[ExpressionToken, Expression]] = []
        for token in self.get_all_tokens():
            if not isinstance(token, TableColumnReferenceToken):
                new_tokens.append(token)
                continue

            if table_name and table_name != token.table_name:
                raise Exception('expressions with different table references are not allowed.')

            table_name = token.table_name if token.table_name and not table_name else table_name
            column_name = column_name or token.column_name
            new_tokens.extend(Expression.column_reference(token.column_name).data)

        return table_name, column_name, Expression(new_tokens)

    def get_references(self) -> Dict[str, 'BachSqlModel']:
        rv = {}
        for data_item in self.data:
            if isinstance(data_item, Expression):
                rv.update(data_item.get_references())
            elif isinstance(data_item, ModelReferenceToken):
                rv[data_item.refname()] = data_item.model
        return rv

    def get_all_tokens(self) -> List[ExpressionToken]:
        result = []
        for data_item in self.data:
            if isinstance(data_item, Expression):
                result.extend(data_item.get_all_tokens())
            else:
                result.append(data_item)
        return result

    def to_sql(self, dialect: Dialect, table_name: Optional[str] = None) -> str:
        """
        Compile the expression to a SQL fragment by calling to_sql() on every token or expression in data
        :param table_name: Optional table name, if set all column-references will be compiled as
            '"{table_name}"."{column_name}"' instead of just '"{column_name}"'.
        :return SQL representation of the expression.
        """
        resolved_tables_expression = self.resolve_column_references(dialect, table_name)
        return ''.join(
            d.to_sql(dialect=dialect) for d in resolved_tables_expression.data
        )


class NonAtomicExpression(Expression):
    """
    An expression that needs '( .. )' around it when used together with other Expressions
    in Expression.construct(). This subclass is required, because not all Expressions need to be wrapped
    in parenthesis, e.g. `expression` `<` `ANY (subquery)` should probably have `expression` wrapped if it's
    a complex expression, but not the `ANY ...` part, since that would not be valid SQL.
    """
    pass


class IndependentSubqueryExpression(Expression):
    pass


class SingleValueExpression(Expression):
    """
    An Expression that is expected to return just one value.
    If wrapped around IndependentSubqueryExpression, this will still have is_independent_subquery == True
    """
    @property
    def is_independent_subquery(self) -> bool:
        # If this Expression is wrapped around a IndependentSubqueryExpression, most likely, there will be
        # just one in here, but let's make sure.
        all_isq = [d.is_independent_subquery for d in self._data if isinstance(d, Expression)]
        return len(all_isq) > 0 and all(all_isq)


class ConstValueExpression(SingleValueExpression):
    pass


class AggregateFunctionExpression(Expression):
    @property
    def is_constant(self) -> bool:
        # We don't consider an aggregate function constant even if all its subexpressions are,
        # because it requires materialization of the aggregation to actually be a constant value again.
        # Maybe we will revisit this some day. If that day comes, make sure to look at Series.count() as well.
        return False


class WindowFunctionExpression(Expression):
    """
    A WindowFunctionExpression contains an aggregation- or window function, and a window clause:
    e.g. agg_func() OVER (...). The agg_func. It's not a subclass of AggregateFunctionExpression because
    a window expression makes sense without the main query having a GROUP BY clause, as the partitioning
    is contained within the expression.

    """
    @property
    def is_constant(self) -> bool:
        # We don't consider an window expression constant even if all its subexpressions are,
        # because it requires materialization to actually be a constant value again.
        # Maybe we will revisit this some day. If that day comes, make sure to look at Series.count() as well.
        return False

    @property
    def has_aggregate_function(self) -> bool:
        # If a window expression contains an aggregate function, it's not an aggregate expression
        return False


class MultiLevelExpression(Expression):
    """
    A MultiLevelExpression contains multiple expressions referencing to different columns.
    """
    def to_sql(self, dialect: Dialect, table_name: Optional[str] = None) -> str:
        # same as original function, we just need to join by a comma since parent expression is composed
        # by multiple column references
        resolved_tables_expression = self.resolve_column_references(dialect, table_name)
        return ','.join(
            d.to_sql(dialect=dialect) for d in resolved_tables_expression.data
        )


def join_expressions(expressions: Sequence[Expression], join_str: str = ', ') -> Expression:
    """
    Construct an expression consisting of the given list of expressions joined by the join_str.
    """
    fmt = join_str.join(['{}'] * len(expressions))
    return Expression.construct(fmt, *expressions)


def get_variable_tokens(expressions: List['Expression']) -> Set[VariableToken]:
    """
    Get the names of all VariableTokens in the list of expressions
    """
    found_tokens = set()
    for expression in expressions:
        for token in expression.get_all_tokens():
            if isinstance(token, VariableToken):
                found_tokens.add(token)
    return found_tokens



"""
Copyright 2021 Objectiv B.V.
"""
from typing import List, NamedTuple, Dict, Set, Iterable

from sqlalchemy.engine import Dialect

from sql_models.graph_operations import find_nodes, FoundNode
from sql_models.model import SqlModel, REFERENCE_UNIQUE_FIELD, Materialization
from sql_models.sql_query_parser import raw_sql_to_selects
from sql_models.util import quote_identifier, is_postgres, is_bigquery, DatabaseNotSupportedException


class GeneratedSqlStatement(NamedTuple):
    name: str
    materialization: Materialization
    sql: str


def to_sql(dialect: Dialect, model: SqlModel) -> str:
    """
    Give the sql for the given SqlModel.

    The returned sql can be a single query or statement, or can contain multiple sql statements. The latter
    will be the case if the dependency graph of model contains other models with a materialization that is a
    statement but has non-lasting effect (e.g. creating temporary tables). Then the returned SQL will also
    contain statements for those models.

    If the dependency graph contains models with lasting effect (e.g. create a table or view), then no
    statements for those dependencies are included in the returned sql. The returned sql will assume those
    dependencies have been created already.

    The generated SQL always includes the model itself, regardless of whether it has lasting effect or not.
    Examples: If model.materialization.type_name == 'table' then the final statement of the returned sql will
    create a table. If the materialization is 'query' then the final statement of the returned sql will
    execute a query.

    :param dialect: SQL Dialect
    :param model: model to convert to sql
    :return: executable SQL statement
    """
    # Find all nodes that are a statement (e.g. create a temporary table), but exclude nodes that have a
    # lasting effect (e.g. creating a regular table). Always include the start node
    # Make sure we get the longest possible path to a node (use_last_found_instance=True). That way we can
    # reverse the list and we'll get the nodes that are a dependency for other nodes before the node that
    # depends on them.
    materialized_found_nodes: List[FoundNode] = find_nodes(
        start_node=model,
        function=lambda node: (
            (node is model)
            or
            (node.materialization.is_statement and not node.materialization.has_lasting_effect)
        ),
        first_instance=False
    )

    models = [fn.model for fn in reversed(materialized_found_nodes)]
    statements = _to_sql_list_models(dialect=dialect, models=models)
    sql = ';\n'.join(sql_stat.sql for sql_stat in statements)
    return sql


def to_sql_materialized_nodes(
        dialect: Dialect,
        start_node: SqlModel,
        include_start_node=True,
) -> List[GeneratedSqlStatement]:
    """
    Give list of sql statements:
        * The sql to query the given model
        * The sql to create all views and tables that the given model depends upon
    :param dialect: SQL Dialect
    :param start_node: model to convert to sql
    :return: A list of generated sql statements. Each object contains the name of the model, the
        materialization of the model, and the sql to run the query, or create the table, or create the view.
        The order of the items is significant: earlier statements will create views and/or tables that might
        be used by later statements.
        The names are guaranteed to be unique in the list.
    """
    # find all nodes that are materialized as view or table, and the start_node if needed
    # make sure we get the longest possible path to a node (use_last_found_instance=True). That way we can
    # reverse the list and we'll get the nodes that are a dependency for other nodes before the node that
    # depends on them.

    materialized_found_nodes: List[FoundNode] = find_nodes(
        start_node=start_node,
        function=lambda node: (
            (node is start_node and include_start_node) or node.materialization.is_statement
        ),
        first_instance=False
    )
    models = [fn.model for fn in reversed(materialized_found_nodes)]
    return _to_sql_list_models(dialect=dialect, models=models)


def _to_sql_list_models(dialect: Dialect, models: List[SqlModel]) -> List[GeneratedSqlStatement]:
    """

    """
    result: List[GeneratedSqlStatement] = []
    compiler_cache: Dict[str, List['SemiCompiledTuple']] = {}
    _check_names_unique(dialect, models)
    for model in models:
        generated_sql = GeneratedSqlStatement(
            name=model_to_name(dialect, model),
            sql=_to_sql_materialized_node(
                dialect=dialect,
                model=model,
                compiler_cache=compiler_cache),
            materialization=model.materialization
        )
        result.append(generated_sql)
    return result


def _to_sql_materialized_node(
        dialect: Dialect,
        model: SqlModel,
        compiler_cache: Dict[str, List['SemiCompiledTuple']],
) -> str:
    """
    Give the sql to query the given model
    :param dialect: SQL Dialect
    :param model: model to convert to sql
    :param compiler_cache: Dictionary mapping model hashes to already compiled results
    :return: executable select query
    """
    queries = _to_cte_sql(dialect=dialect, compiler_cache=compiler_cache, model=model)
    queries = _filter_duplicate_ctes(queries)
    if len(queries) == 0:
        # _to_cte_sql only returns an empty list when there is a source node,
        # but it's not referenced. We can't generate anything in that scenario.
        raise Exception('No models to compile')

    if len(queries) == 1:
        return _materialize(dialect=dialect, sql_query=queries[0].sql, model=model)

    # case: len(result) > 1
    sql = 'with '
    sql += ',\n'.join(f'{row.quoted_cte_name} as ({row.sql})' for row in queries[:-1])
    sql += '\n' + queries[-1].sql
    return _materialize(dialect=dialect, sql_query=sql, model=model)


def _materialize(dialect: Dialect, sql_query: str, model: SqlModel) -> str:
    """
    Generate sql that wraps the sql_query with the materialization indicated by model.
    :param dialect: SQL Dialect
    :param sql_query: raw sql query
    :param model: model that indicates the materialization and name of the resulting view or table
        (if applicable).
    :return: raw sql
    """

    materialization = model.materialization
    quoted_name = model_to_quoted_name(dialect, model)
    if materialization in [Materialization.CTE, Materialization.QUERY]:
        return sql_query
    if materialization == Materialization.VIEW:
        return f'create view {quoted_name} as {sql_query}'
    if materialization == Materialization.TABLE:
        return f'create table {quoted_name} as {sql_query}'
    if materialization == Materialization.TEMP_TABLE:
        if is_postgres(dialect):
            return f'create temporary table {quoted_name} on commit drop as {sql_query}'
        if is_bigquery(dialect):
            return f'CREATE TEMP TABLE {quoted_name} AS {sql_query}'
        raise DatabaseNotSupportedException(dialect, f'Cannot create temporary table for {dialect.name}')
    if materialization in [Materialization.VIRTUAL_NODE, Materialization.SOURCE]:
        return ''
    raise Exception(f'Unsupported Materialization value: {materialization}')


class SemiCompiledTuple(NamedTuple):
    """
    Object representing a single CTE select statement from a big select statement
    with common table expressions.
    """
    # This is very similar to the CteTuple in sql_query_parser. However here cte_name is mandatory and
    # quoted and escaped.
    quoted_cte_name: str
    sql: str


def _check_names_unique(dialect: Dialect, models: Iterable[SqlModel]):
    """
    Check that there are no duplicate names in the list of models. Raises an error if duplicates are found.
    """
    seen: Set[str] = set()
    for model in models:
        name = model_to_name(dialect, model)
        if name in seen:
            raise ValueError(f'Names of SqlModels need to be unique throughout the graph.'
                             f'Duplicate found: "{name}"')
        seen.add(name)


def _filter_duplicate_ctes(queries: List[SemiCompiledTuple]) -> List[SemiCompiledTuple]:
    """
    Filter duplicate CTEs from the list
    If a cte occurs multiple times, then only keep the first occurrence.
    Throw an error if not all of the occurrences are the same.
    :param queries:
    :return:
    """
    seen: Dict[str, str] = {}
    result = []
    for query in queries:
        if query.quoted_cte_name not in seen:
            seen[query.quoted_cte_name] = query.sql
            result.append(query)
        elif seen[query.quoted_cte_name] != query.sql:
            raise Exception(f'Encountered the CTE {query.quoted_cte_name} multiple times, but with different '
                            f'definitions. HINT: use "{{REFERENCE_UNIQUE_FIELD}}" in the sql definition '
                            f'to make CTE names unique between different instances of the same model.\n'
                            f'first: {seen[query.quoted_cte_name]}\n'
                            f'second: {query.sql}\n')
    return result


def _to_cte_sql(dialect: Dialect,
                compiler_cache: Dict[str, List[SemiCompiledTuple]],
                model: SqlModel
                ) -> List[SemiCompiledTuple]:
    """
    Recursively build the list of all common table expressions that are needed to generate the sql for
    the given model
    :param dialect: SQL Dialect
    :param compiler_cache: Dictionary mapping model hashes to already compiled results
    :param model: model to convert to a list of SemiCompiledTuple
    :return:
    """
    if model.hash in compiler_cache:
        return compiler_cache[model.hash]

    # First recursively compile all CTEs that we depend on
    result = []
    reference_names = {
        name: model_to_quoted_name(dialect=dialect, model=reference)
        for name, reference in model.references.items()
    }
    for ref_name, reference in model.references.items():
        if reference.materialization.is_cte:
            result.extend(_to_cte_sql(dialect=dialect, compiler_cache=compiler_cache, model=reference))

    # Compile the actual model
    result.extend(
        _single_model_to_sql(
            dialect=dialect,
            compiler_cache=compiler_cache,
            model=model,
            reference_names=reference_names
        )
    )

    compiler_cache[model.hash] = result
    return result


def model_to_name(dialect: Dialect, model: SqlModel):
    """
    Get the name for the cte/table/view that will be generated from this model, quoted and escaped.
    """
    if model.materialization == Materialization.SOURCE:
        if model.materialization_name is None:
            raise Exception('SqlModel with SOURCE materialization must have materialization_name set.')
        return model.materialization_name

    # max length of an identifier name in Postgres is normally 63 characters.
    # We'll use that as a cutoff if we don't know better
    if is_bigquery(dialect):
        max_length = 1023
    else:
        max_length = 63

    if model.materialization_name is not None:
        return model.materialization_name[0:max_length]
    name = f'{model.generic_name[0:(max_length-35)]}___{model.hash}'
    return name


def model_to_quoted_name(dialect: Dialect, model: SqlModel):
    """
    Get the name for the cte/table/view that will be generated from this model, quoted and escaped.
    """
    return quote_identifier(dialect, model_to_name(dialect, model))


def _single_model_to_sql(dialect: Dialect,
                         compiler_cache: Dict[str, List[SemiCompiledTuple]],
                         model: SqlModel,
                         reference_names: Dict[str, str]
                         ) -> List[SemiCompiledTuple]:
    """
    Split the sql for a given model into a list of separate CTEs.
    :param dialect: SQL Dialect
    :param compiler_cache: Dictionary mapping model hashes to already compiled results
    :param model:
    :param reference_names: mapping of references in the raw sql, to the names of the CTEs that they refer
    :return:
    """
    if model.hash in compiler_cache:
        return compiler_cache[model.hash]

    result: List[SemiCompiledTuple] = []
    sql = model.sql
    if sql == '':
        # If the SQL is completely empty, this is the first node that we're selecting from.
        # We only need to store it in order to select from the materialization name
        compiler_cache[model.hash] = result
        return result

    # If there are any format strings in the placeholder values that need escaping, they should have been
    # escaped by now.
    # Otherwise, this will cause trouble the next time we call format() below for the references
    sql = _format_sql(sql=sql, values=model.placeholders_formatted, model=model)
    # {{id}} (==REFERENCE_UNIQUE_FIELD) is a special placeholder that gets the unique model identifier,
    # which can be used in templates to make sure that if a model gets used multiple times,
    # the cte-names are still unique.
    _reference_names = dict(**reference_names)
    _reference_names[REFERENCE_UNIQUE_FIELD] = model.hash
    sql = _format_sql(sql=sql, values=_reference_names, model=model)
    ctes = raw_sql_to_selects(sql)

    for cte in ctes[:-1]:
        # For all CTEs the name should be set. Only for the final select (== cte[-1]) it will be None.
        assert cte.name is not None
        result.append(
            SemiCompiledTuple(quoted_cte_name=quote_identifier(dialect, cte.name), sql=cte.select_sql)
        )
    result.append(
        SemiCompiledTuple(quoted_cte_name=model_to_quoted_name(dialect, model), sql=ctes[-1].select_sql)
    )

    compiler_cache[model.hash] = result
    return result


def _format_sql(sql: str, values: Dict[str, str], model: SqlModel):
    """ Execute sql.format(**values), and if that fails raise a clear exception. """
    try:
        sql = sql.format(**values)
    except Exception as exc:
        raise Exception(f'Failed to format sql for model {model.generic_name}. \n'
                        f'Format values: {values}. \n'
                        f'Sql: {sql}') from exc
    return sql



from abc import ABC, abstractmethod
from typing import TypeVar, Dict, Tuple, cast

from bach.expression import Expression, join_expressions
from bach.sql_model import BachSqlModel, construct_references
from sql_models.model import CustomSqlModelBuilder, Materialization

from bach.series import SeriesJson, SeriesInt64
from bach.utils import get_name_to_column_mapping

TSeriesJson = TypeVar('TSeriesJson', bound='SeriesJson')

_ITEM_IDENTIFIER_EXPR = Expression.identifier(name='__unnest_item')
_OFFSET_IDENTIFIER_EXPR = Expression.identifier(name='__unnest_item_offset')


class ArrayFlattening(ABC):
    """
    Abstract class that expands an array-type column into a set of rows.

    Child classes are in charge of
    1) specifying the correct expressions for unnesting arrays, by overriding `_get_cross_join_expression`.
    2) giving the correct offset for each item, by setting `_db_offset_is_one_based`.

    .. note::
        Final result will always have different base node than provided Series object.

    returns Tuple with:
        - SeriesJson: Representing the element of the array.
        - SeriesInt64: Offset of the element in the array
    """

    """
    _db_offset_is_one_based can be overridden by child classes to indicate that the offsets from the database
    are 1-based offsets instead of 0-based offsets.
    """
    _db_offset_is_one_based = False

    def __init__(self, series_object: 'TSeriesJson'):
        self._series_object = series_object.copy()

        if not self._series_object.is_materialized:
            self._series_object = cast(
                TSeriesJson, self._series_object.materialize(node_name='array_flatten')
            )

    def __call__(self, *args, **kwargs) -> Tuple['TSeriesJson', 'SeriesInt64']:
        from bach import DataFrame, SeriesInt64

        dialect = self._series_object.engine.dialect
        name_to_column_mapping = get_name_to_column_mapping(dialect=dialect, names=self.all_dtypes.keys())

        unnested_array_df = DataFrame.from_model(
            engine=self._series_object.engine,
            model=self._get_unnest_model(),
            index=list(self._series_object.index.keys()),
            all_dtypes=self.all_dtypes,
            name_to_column_mapping=name_to_column_mapping
        )
        item_series = unnested_array_df[self.item_series_name]
        offset_series = unnested_array_df[self.item_offset_series_name]
        return (
            cast(TSeriesJson, item_series),
            offset_series.copy_override_type(SeriesInt64),
        )

    @property
    def item_series_name(self) -> str:
        """
        Final name of the series containing the array elements.
        """
        return self._series_object.name

    @property
    def item_offset_series_name(self) -> str:
        """
        Final name of the series containing the offset of the element in the array.
        """
        return f'{self._series_object.name}_offset'

    @property
    def all_dtypes(self) -> Dict[str, str]:
        """
        Mapping of all dtypes of all referenced columns in generated model
        """
        return {
            **{idx.name: idx.dtype for idx in self._series_object.index.values()},
            self.item_series_name: self._series_object.dtype,
            self.item_offset_series_name: 'int64'
        }

    def _get_unnest_model(self) -> BachSqlModel:
        """
        Creates a BachSqlModel in charge of expanding the array column.
        """
        column_expressions = self._get_column_expressions()
        select_column_expr = join_expressions(list(column_expressions.values()))
        from_model_expr = Expression.model_reference(self._series_object.base_node)
        cross_join_expr = self._get_cross_join_expression()

        sql_exprs = [select_column_expr, from_model_expr, cross_join_expr]
        dialect = self._series_object.engine.dialect
        sql = Expression.construct('SELECT {} FROM {} CROSS JOIN {}', *sql_exprs).to_sql(dialect)

        return BachSqlModel(
            model_spec=CustomSqlModelBuilder(sql=sql, name='unnest_array'),
            placeholders={},
            references=construct_references(base_references={}, expressions=sql_exprs),
            materialization=Materialization.CTE,
            materialization_name=None,
            column_expressions=column_expressions,
        )

    def _get_column_expressions(self) -> Dict[str, Expression]:
        """
        Final column expressions for the generated model
        """
        dialect = self._series_object.engine.dialect

        offset_expr = _OFFSET_IDENTIFIER_EXPR
        if self._db_offset_is_one_based:
            offset_expr = Expression.construct('{} - 1', offset_expr)  # Normalize to 0-based offset

        return {
            **{
                idx.name: idx.expression for idx in self._series_object.index.values()
            },
            self.item_series_name: Expression.construct_expr_as_sql_name(
                dialect=dialect, expr=_ITEM_IDENTIFIER_EXPR, name=self.item_series_name
            ),
            self.item_offset_series_name: Expression.construct_expr_as_sql_name(
                dialect=dialect, expr=offset_expr, name=self.item_offset_series_name,
            ),
        }

    @abstractmethod
    def _get_cross_join_expression(self) -> Expression:
        """
        Expression that unnest/extract elements and offsets from array column. Later used
        in a cross join operation for joining the new set of rows back to the source.
        """
        raise NotImplementedError()


class BigQueryArrayFlattening(ArrayFlattening):

    # BigQuery uses 0-based offsets.
    _db_offset_is_one_based = False

    def _get_cross_join_expression(self) -> Expression:
        """ For documentation, see implementation in class :class:`ArrayFlattening` """
        return Expression.construct(
            'UNNEST(JSON_QUERY_ARRAY({}.{})) AS {} WITH OFFSET AS {}',
            Expression.model_reference(self._series_object.base_node),
            self._series_object,
            _ITEM_IDENTIFIER_EXPR,
            _OFFSET_IDENTIFIER_EXPR,
        )


class PostgresArrayFlattening(ArrayFlattening):

    # Postgres uses 1-based ordinality instead of 0-based offsets.
    _db_offset_is_one_based = True

    def _get_cross_join_expression(self) -> Expression:
        """ For documentation, see implementation in class :class:`ArrayFlattening` """
        return Expression.construct(
            'JSONB_ARRAY_ELEMENTS({}.{}) WITH ORDINALITY AS _unnested({}, {})',
            Expression.model_reference(self._series_object.base_node),
            self._series_object,
            _ITEM_IDENTIFIER_EXPR,
            _OFFSET_IDENTIFIER_EXPR,
        )


class AthenaArrayFlattening(ArrayFlattening):

    # Athena uses 1-based ordinality instead of 0-based offsets.
    _db_offset_is_one_based = True

    def _get_cross_join_expression(self) -> Expression:
        """ For documentation, see implementation in class :class:`ArrayFlattening` """
        return Expression.construct(
            'unnest(cast({}.{} as array(json))) with ordinality as _unnested({}, {})',
            Expression.model_reference(self._series_object.base_node),
            self._series_object,
            _ITEM_IDENTIFIER_EXPR,
            _OFFSET_IDENTIFIER_EXPR,
        )



"""
Copyright 2021 Objectiv B.V.

All information is in setup.cfg. But it seems 'pip install -e' likes to have a setup.py, so we have this
file. We might need to look more in into the whole packaging thing in the future.
"""

from setuptools import setup

if __name__ == "__main__":
    setup()



"""
Copyright 2022 Objectiv B.V.
"""
import operator
from abc import ABC, abstractmethod
from copy import copy
from functools import reduce
from typing import Union, List, Optional, Dict, Any, cast, TypeVar, Tuple, Type, TYPE_CHECKING

from bach import DataFrameOrSeries, SortColumn
from bach.series.series import Series


import pandas
from sqlalchemy.engine import Engine, Dialect
from bach.expression import Expression, MultiLevelExpression
from bach.series.series import ToPandasInfo
from bach.sql_model import BachSqlModel
from bach.types import AllSupportedLiteralTypes, get_series_type_from_dtype, StructuredDtype
from sql_models.constants import DBDialect, NotSet, not_set
from sql_models.util import DatabaseNotSupportedException, is_postgres, is_bigquery, is_athena

T = TypeVar('T', bound='SeriesAbstractMultiLevel')

if TYPE_CHECKING:
    from bach.partitioning import GroupBy
    from bach.series import SeriesBoolean, SeriesAbstractNumeric, SeriesFloat64, SeriesInt64, SeriesString
    from bach.dataframe import DataFrame

_BOUNDS_PANDAS_INTERVAL_CLOSED_MAPPING = {
    '[]': 'both',
    '[)': 'left',
    '(]': 'right',
    '()': 'neither',
}


class SeriesAbstractMultiLevel(Series, ABC):
    """
    A Series that represent the base multi-level types. Works as a data structure containing multiple series
    of different dtypes (based on the structure defined by the child class).

    Each level is treated as an individual series till user calls the following methods:
        * `SeriesAbstractMultiLevel.to_pandas`
        * `SeriesAbstractMultiLevel.to_numpy`
        * `SeriesAbstractMultiLevel.view_sql`

    Each child class must be in charge of:
        * having each level as a property (ensuring immutability)
        * a column expression representing the final structure to be built.

    ** Operations **
        Adapts all supported methods by Series class. Currently, aggregations are not supported.
    """
    def copy_override(
        self: T,
        *,
        engine: Optional[Engine] = None,
        base_node: Optional[BachSqlModel] = None,
        index: Optional[Dict[str, 'Series']] = None,
        name: Optional[str] = None,
        expression: Optional['Expression'] = None,
        group_by: Optional[Union['GroupBy', NotSet]] = not_set,
        instance_dtype: Optional[StructuredDtype] = None,
        order_by: Optional[List[SortColumn]] = None,
        **kwargs,
    ) -> T:
        """
        INTERNAL: Copy this instance into a new one, with the given overrides

        SeriesAbstractMultiLevel only provides current levels if not present in extra arguments
        """
        extra_params = copy(self.levels)
        extra_params.update(kwargs)
        return cast(T, super().copy_override(
            engine=engine,
            base_node=base_node,
            index=index,
            name=name,
            expression=expression,
            group_by=group_by,
            instance_dtype=instance_dtype,
            order_by=order_by,
            **extra_params
        ))

    @classmethod
    def get_class_instance(
        cls,
        engine: Engine,
        base_node: BachSqlModel,
        index: Dict[str, 'Series'],
        name: str,
        expression: Expression,
        group_by: Optional['GroupBy'],
        order_by: Optional[List[SortColumn]],
        instance_dtype: StructuredDtype,
        **kwargs
    ):
        """
        INTERNAL: Create an instance of this class.

        Creates a new instance for each level series.
        """
        base_params = {
            'engine': engine,
            'base_node': base_node,
            'group_by': group_by,
            'index': index,
            'order_by': order_by,
        }

        default_level_dtypes = cls.get_supported_level_dtypes()

        # check if all level are provided in kwargs and each of them is a series instance
        levels_are_provided = all(
            level in kwargs and isinstance(kwargs[level], Series)
            for level in default_level_dtypes.keys()
        )

        if not levels_are_provided:
            # if levels are not provided, search for them in base node
            # this can only happen when instantiating a new DataFrame, if df is containing a multilevel series
            # then levels MUST be in the base node.
            missing_references = [
                f'_{name}_{level_name}' for level_name in default_level_dtypes.keys()
                if f'_{name}_{level_name}' not in base_node.series_names
            ]
            if missing_references:
                raise ValueError(
                    f'base node must include all referenced columns {missing_references} '
                    f'for the "{cls.__name__}" instance.'
                )

        sub_levels = {}
        for level_name in default_level_dtypes.keys():
            if levels_are_provided:
                # caller should be responsible for providing correct column references
                level_base = kwargs[level_name]
                dtype = level_base.dtype
                expr = level_base.expression
            else:
                dtype = default_level_dtypes[level_name][0]
                expr = Expression.column_reference(f'_{name}_{level_name}')

            sub_levels[level_name] = get_series_type_from_dtype(dtype).get_class_instance(
                name=f'_{name}_{level_name}',
                expression=expr,
                instance_dtype=dtype,
                **base_params
            )

        return cls(
            engine=engine,
            base_node=base_node,
            index=index,
            name=name,
            expression=expression,
            group_by=group_by,
            order_by=order_by,
            instance_dtype=instance_dtype,
            **sub_levels
        )

    @property
    def levels(self) -> Dict[str, Series]:
        """returns mapping between level name and series"""
        return {
            attr: getattr(self, attr) for attr in self.get_supported_level_dtypes().keys()
        }

    @property
    def expression(self) -> Expression:
        """ INTERNAL: Get the MultiLevelExpression composed by all levels' expression"""
        return MultiLevelExpression([level.expression for level in self.levels.values()])

    @property
    def level_expressions(self) -> List[Expression]:
        """ INTERNAL: Returns a list with each level expression"""
        return [level.expression for level in self.levels.values()]

    @abstractmethod
    def get_column_expression(self, table_alias: str = None) -> Expression:
        """
        INTERNAL: Child class must be in charge of providing the correct expression that represents the type.
        """
        raise NotImplementedError()

    @classmethod
    def get_supported_level_dtypes(cls) -> Dict[str, Tuple[str, ...]]:
        """returns mapping of supported dtypes per level"""
        raise NotImplementedError()

    @classmethod
    def from_value(
        cls: Type[T],
        base: DataFrameOrSeries,
        value: Any,
        name: str = 'new_series',
        dtype: Optional[StructuredDtype] = None
    ) -> T:
        """
        Create an instance of this class, that represents a column with the given value.
        The returned Series will be similar to the Series given as base. In case a DataFrame is given,
        it can be used immediately with that frame.
        :param base:    The DataFrame or Series that the internal parameters are taken from
        :param value:   None or a mapping between each level and constant. All levels must be present.
        :param name:    The name that it will be known by (only for representation)
        """
        if dtype is None:
            dtype = cls.dtype
        if (
            value is not None
            and (
                not isinstance(value, dict)
                or not all(level in value for level in cls.get_supported_level_dtypes().keys())
            )
        ):
            raise ValueError(f'value should contain mapping for each {cls.__name__} level')

        from bach.series.series import value_to_series

        if value is not None:
            levels = {
                level_name: value_to_series(base=base, value=level_value)
                for level_name, level_value in value.items()
            }
        else:
            levels = {
                level_name: get_series_type_from_dtype(dtypes[0]).from_value(
                    base=base, value=None, name=level_name)
                for level_name, dtypes in cls.get_supported_level_dtypes().items()
            }
        result = cls.get_class_instance(
            engine=base.engine,
            base_node=base.base_node,
            index=base.index,
            name=name,
            expression=Expression.construct(''),
            group_by=None,
            order_by=[],
            instance_dtype=dtype,
            **levels
        )

        return result

    @classmethod
    def dtype_to_expression(cls, dialect: Dialect, source_dtype: str, expression: Expression) -> Expression:
        raise NotImplementedError()

    @classmethod
    def supported_literal_to_expression(cls, dialect: Dialect, literal: Expression) -> Expression:
        raise NotImplementedError()

    @classmethod
    def supported_value_to_literal(cls, dialect: Dialect, value: Any, dtype: StructuredDtype) -> Expression:
        raise NotImplementedError()

    def astype(self, dtype: Union[str, Type]) -> 'Series':
        raise NotImplementedError(f'{self.__class__.__name__} cannot be casted to other types.')

    def equals(self, other: Any, recursion: str = None) -> bool:
        """
        Checks if all levels between both multi-level series are the same.

        For more information look at :py:meth:`bach.Series.equals()`
        """
        if not isinstance(other, self.__class__):
            return False

        same_levels = all(
            self.levels[level_name].equals(other.levels[level_name], recursion)
            for level_name in self.levels.keys()
        )
        return same_levels and super().equals(other, recursion)

    def isnull(self) -> 'SeriesBoolean':
        """
        Checks if all levels are null.

        For more information look at :py:meth:`bach.Series.isnull()`
        """
        all_series = [lvl.isnull() for lvl in self.levels.values()]
        return reduce(operator.and_, all_series)

    def notnull(self) -> 'SeriesBoolean':
        """
        Checks if all levels are not null.

        For more information look at :py:meth:`bach.Series.notnull()`
        """
        all_series = [lvl.notnull() for lvl in self.levels.values()]
        return reduce(operator.and_, all_series)

    def fillna(self, other: AllSupportedLiteralTypes):
        """
        Fill any NULL value of any or all levels.
        For more information look at :py:meth:`bach.Series.fillna()`

        ..note:
            "other" should be a dict object containing at least one level to be filled.
        """
        if (
            not isinstance(other, dict)
            or not all(level in self.levels for level in other.keys())
        ):
            raise ValueError(
                f'"other" should contain mapping for at least one of {self.__class__.__name__} levels'
            )

        self_cp = self.copy()
        for level_name, value in other.items():
            setattr(self_cp, f'_{level_name}', self.levels[level_name].fillna(value))

        return self_cp

    def append(
        self,
        other: Union['Series', List['Series']],
        ignore_index: bool = False,
    ) -> 'Series':
        """
        Append rows of other multi-level series to the caller series.
        For more information look at :py:meth:`bach.Series.append()`

        ..note:
            All elements to be appended must be the same multi-level dtype as the caller.
        """
        if not other:
            return self

        levels_df_to_append = []
        for series in other if isinstance(other, list) else [other]:
            if not isinstance(series, self.__class__):
                raise ValueError(f'can only append "{self.dtype}" series to {self.name}')

            level_df = series.copy_override(name=self.name).flatten()
            levels_df_to_append.append(level_df)

        appended_df = self.flatten().append(levels_df_to_append)
        return self.from_value(
            base=appended_df,
            value={
                level_name: appended_df[f'_{self.name}_{level_name}'] for level_name in self.levels.keys()
            },
            name=self.name,
            dtype=self.dtype
        )

    def flatten(self) -> 'DataFrame':
        """
        Converts the multi-level series into a DataFrame,
        where each level is a series of the resultant DataFrame.
        """
        from bach.dataframe import DataFrame
        from bach.savepoints import Savepoints
        return DataFrame(
            engine=self.engine,
            base_node=self.base_node,
            index=self.index,
            series={level.name: level for level in self.levels.values()},
            group_by=self.group_by,
            order_by=[],
            savepoints=Savepoints(),
            variables={},
        )

    def _parse_level_value(self, level_name: str, value: Union[AllSupportedLiteralTypes, Series]) -> 'Series':
        """
        INTERNAL: validates that the level value is supported by the caller.
        """
        supported_dtypes = self.get_supported_level_dtypes()
        if level_name not in supported_dtypes:
            raise ValueError(f'{level_name} is not a supported level in {self.__class__.__name__}.')

        from bach.series.series import value_to_series
        level = value_to_series(base=self, value=value)

        if not any(
            isinstance(level, get_series_type_from_dtype(dtype)) for dtype in supported_dtypes[level_name]
        ):
            raise ValueError(f'"{level_name}" level should be any of {supported_dtypes[level_name]} dtypes.')

        # level should have same attributes as parent
        return level.copy_override(
            name=f'_{self.name}_{level_name}',
            engine=self.engine,
            base_node=self.base_node,
            index=self.index,
            group_by=self.group_by,
            order_by=self.order_by,
        )


class SeriesNumericInterval(SeriesAbstractMultiLevel):
    """
    Multi-Level Series representing a structure for numeric intervals. This is due that not all supported
    databases have a type that emulates a similar behavior as pandas.Interval dtype.

    For each supported database, structure is composed by 3 levels:
        * `lower`: a numerical value representing the lowest edge of the interval
        * `upper`: a numerical value representing the upper edge of the interval
        * `bounds`: a string value representing the boundary type of the interval:
            Supported values are: '[]', '[)' and '()', if value is not supported,
            then '()' will be used as default when parsing results into `pandas.DataFrame`

    DB representation per engine:
        * postgres: NUMRANGE type will be built as a result.
            For more information: https://www.postgresql.org/docs/current/rangetypes.html
        * bigquery: BigQuery has no direct type that handles ranges, therefore a STRUCT type column is used as
            a final representation.
    """
    dtype = 'numeric_interval'
    dtype_aliases = (pandas.Interval, 'numrange')
    supported_db_dtype = {
        DBDialect.POSTGRES: 'numrange',
        DBDialect.BIGQUERY: 'STRUCT<lower FLOAT64, upper FLOAT64, bounds STRING(2)>',
        DBDialect.ATHENA: 'STRUCT<lower: double, upper: double, bounds: varchar(2)>',
    }

    supported_value_types = (pandas.Interval, dict)

    def __init__(
        self,
        *,
        lower: Union['SeriesAbstractNumeric', float, int],
        upper: Union['SeriesAbstractNumeric', float, int],
        bounds: Union['SeriesString', str],
        **kwargs,
    ) -> None:
        # multi-level series should have an empty expression
        kwargs['expression'] = Expression.construct('')
        super().__init__(**kwargs)

        self._lower = cast('SeriesAbstractNumeric', self._parse_level_value(level_name='lower', value=lower))
        self._upper = cast('SeriesAbstractNumeric', self._parse_level_value(level_name='upper', value=upper))
        self._bounds = cast('SeriesString', self._parse_level_value(level_name='bounds', value=bounds))

    @property
    def lower(self) -> 'SeriesAbstractNumeric':
        return self._lower

    @property
    def upper(self) -> 'SeriesAbstractNumeric':
        return self._upper

    @property
    def bounds(self) -> 'SeriesString':
        return self._bounds

    @classmethod
    def get_supported_level_dtypes(cls) -> Dict[str, Tuple[str, ...]]:
        from bach.series import SeriesFloat64, SeriesInt64, SeriesString
        return {
            'lower': (SeriesFloat64.dtype, SeriesInt64.dtype),
            'upper': (SeriesFloat64.dtype, SeriesInt64.dtype),
            'bounds': (SeriesString.dtype, ),
        }

    def to_pandas_info(self) -> Optional[ToPandasInfo]:
        if is_postgres(self.engine):
            return ToPandasInfo(dtype='object', function=self._parse_numeric_interval_value_postgres)
        if is_bigquery(self.engine):
            return ToPandasInfo(dtype='object', function=self._parse_numeric_interval_value_bigquery)
        if is_athena(self.engine):
            return ToPandasInfo(dtype='object', function=self._parse_numeric_interval_value_athena)
        return None

    def get_column_expression(self, table_alias: str = None) -> Expression:
        # construct final column based on levels
        if is_postgres(self.engine):
            # casting is needed since numrange does not support float64
            # we cast first to text, as casting double precision to numeric might round decimal values
            base_expr_stmt = f'numrange({{}}::text::numeric, {{}}::text::numeric, {{}})'

        elif is_bigquery(self.engine):
            # BigQuery has no proper datatype for numeric intervals,
            # therefore we should represent it as a struct
            base_expr_stmt = f'struct({{}} as `lower`, {{}} as `upper`, {{}} as `bounds`)'
        elif is_athena(self.engine):
            # Athena has no proper datatype for numeric intervals,
            # therefore we should represent it as a "row" structure type and cast it as a json.
            # If we do not cast the row value to json, the engine will return a stringified structure
            # on the following format: "{lower=#, upper=#, bounds= ""}", which is harder to parse.
            # Meanwhile, casting it as a json will result with an array of 3 elements.
            # Explicitly creating the array in the expression is not possible, as Athena does not
            # allow arrays of different db types.
            base_expr_stmt = (
                'cast(cast(row({}, {}, {}) as row(lower double, upper double, bounds varchar(2))) as json)'
            )
        else:
            raise DatabaseNotSupportedException(self.engine)

        # should return null when all levels are null
        expr = Expression.construct(
            f'CASE WHEN {{}} THEN {base_expr_stmt} ELSE NULL END',
            self.notnull(),
            self.lower,
            self.upper,
            self.bounds,
        )
        dialect = self.engine.dialect
        return Expression.construct_expr_as_sql_name(dialect=dialect, expr=expr, name=self.name)

    @staticmethod
    def _parse_numeric_interval_value_postgres(value) -> Optional[pandas.Interval]:
        """
        Helper function that converts Postgres NUMRANGE values into a pandas.Interval object.
        """
        if value is None:
            return value

        if value.lower_inc and value.upper_inc:
            closed = 'both'
        elif value.lower_inc:
            closed = 'left'
        else:
            closed = 'right'
        return pandas.Interval(float(value.lower), float(value.upper), closed=closed)

    @staticmethod
    def _parse_numeric_interval_value_bigquery(value) -> Optional[pandas.Interval]:
        """
        Helper function that converts BigQuery STRUCT values into a pandas.Interval object.
        """
        if value is None:
            return value

        # expects a dict with interval information
        expected_keys = ['lower', 'upper', 'bounds']
        if not isinstance(value, dict) or not all(k in value for k in expected_keys):
            raise ValueError(f'{value} has not the expected structure.')

        closed = _BOUNDS_PANDAS_INTERVAL_CLOSED_MAPPING[value['bounds']]
        return pandas.Interval(float(value['lower']), float(value['upper']), closed=closed)

    @staticmethod
    def _parse_numeric_interval_value_athena(value) -> Optional[pandas.Interval]:
        """
        Helper function that converts Athena JSON (list) values into a pandas.Interval object.
        """
        if value is None:
            return None

        lower, upper, bounds = tuple(value)
        closed = _BOUNDS_PANDAS_INTERVAL_CLOSED_MAPPING[bounds]
        return pandas.Interval(lower, upper, closed=closed)



"""
Copyright 2022 Objectiv B.V.

Functions for parsing data-type strings from BigQuery to our representation for instance-dtypes
"""
from typing import Mapping, List, Tuple

from bach.types import StructuredDtype, Dtype, get_all_db_dtype_to_series
from sql_models.constants import DBDialect


def bq_db_dtype_to_dtype(db_dtype: str) -> StructuredDtype:
    """
    Given a db_dtype as returned by BigQuery, parse this to an instance-dtype.

    Note: We don't yet support Structs with unnamed fields (e.g. 'STRUCT<INT64>' is not supported.

    :param db_dtype: BigQuery db-dtype, e.g. 'STRING', or 'STRUCT<column_name INT64>', etc
    :return: Instance dtype, e.g. 'string', or {'column_name': 'int64'}
    """
    bq_db_dtype_to_series = get_all_db_dtype_to_series()[DBDialect.BIGQUERY]
    scalar_mapping = {db_dtype: series.dtype for db_dtype, series in bq_db_dtype_to_series.items()}

    tokens = _tokenize(db_dtype)
    pos, result = _tokens_to_dtype(tokens=tokens, pos=0, scalar_mapping=scalar_mapping)
    if pos != len(tokens) - 1:
        raise ValueError(f'Unexpected tokens after last parsed tokens. '
                         f'Last parsed tokens position: {pos}, total tokens: {len(tokens)}')
    return result


# Code below is not super clean, and we should have more tests. But we have tests for bq_db_dtype_to_dtype()
# so we can always refactor this in the future. For now it will do fine.

def _tokenize(bq_db_dtype: str):
    """
    Split the type as received from BigQuery in a way that _tokens_to_dtype can handle:
    Split on '<', '>', ',' and spaces between struct field name and type. Discards all whitespace except
    space between struct field name and type.
    """
    tokens = []
    token = []

    i = iter(bq_db_dtype.strip())
    c = next(i, None)
    while c:
        while c and c not in '<>,':
            token.append(c)
            c = next(i, None)

        # process the token
        t = "".join(token).strip()
        token = []
        if len(t):
            if ' ' in t:
                # This is the format '<identifier> <type>'
                identifier_str, type_str, *too_many = t.split(' ')
                if too_many:
                    raise ValueError('Cannot tokenize string: "{t}", expected only one space.')
                tokens.extend([identifier_str, ' ', type_str])
            else:
                tokens.append(t)
        if c:
            tokens.append(str(c))
            c = next(i, None)

    return tokens


def _tokens_to_dtype(
    tokens: List[str],
    pos: int,
    scalar_mapping: Mapping[str, Dtype]
) -> Tuple[int, StructuredDtype]:
    """
    Parse token to Dtype
    :param tokens: list with all tokens
    :poram pos: token position to start parsing
    :param scalar_mapping: Mapping from BigQuery db_dtype to our Dtypes. Should contain all scalar types.
    :return: tuple: position of the last processed token, and parsed dtype
    """
    t = get_token(tokens, pos)
    if t == 'STRUCT':
        # Recursive case: get type of elements in struct
        pos += 1
        assert_token(tokens, pos, '<')
        current = {}
        while (pos + 1) < len(tokens):
            pos += 1
            if get_token(tokens, pos) == '>':
                break
            elif get_token(tokens, pos) == ',':
                continue
            else:
                # parse a sub type
                if get_token(tokens, pos + 1) == ' ':
                    # format: `name TYPE`
                    name = get_token(tokens, pos)
                    pos, sub_dtype = _tokens_to_dtype(tokens, pos=pos + 2, scalar_mapping=scalar_mapping)
                    current[name] = sub_dtype
                else:
                    # format: `TYPE`
                    raise ValueError('We do not support STRUCTS with unnamed fields.')
        assert_token(tokens, pos, '>')
        return pos, current
    if t == 'ARRAY':
        # Recursive case: get type of elements in array
        assert_token(tokens, pos + 1, '<')
        pos, sub_dtype = _tokens_to_dtype(tokens, pos=pos + 2, scalar_mapping=scalar_mapping)
        assert_token(tokens, pos + 1, '>')
        return pos + 1, [sub_dtype]

    if t in scalar_mapping:
        # Base case: simple type
        return pos, scalar_mapping[t]

    raise ValueError(f'Unexpected token: "{t}" on position: {pos}')


def get_token(tokens: List[str], pos: int) -> str:
    if pos >= len(tokens):
        raise ValueError(f'Expected token on position {pos}, but found no token. '
                         f'len(tokens) = {len(tokens)}. Tokens: {tokens}')
    return tokens[pos]


def assert_token(tokens: List[str], pos: int, expected: str):
    """ Assert that character on pos in tokens is the expected value, if not raise an ValueError. """
    if pos >= len(tokens):
        raise ValueError(f'Expected token "{expected}" on position {pos}, but found no token. '
                         f'len(tokens) = {len(tokens)}. Tokens: {tokens}')
    if tokens[pos] != expected:
        raise ValueError(f'Expected token "{expected}" on position {pos}, but found: {tokens[pos]}. '
                         f'Tokens: {tokens}')
